#!/usr/bin/python

from tkinter import *
from tkinter import ttk
import networkx as nx

import sys

from .pace import read_graph

CPAD = 5
RPAD = 20
LDIST = 200
HDIST = 0
RECT_SIZE = 25

w = 2

graph = None
order = {}
order_reverse = {}

def resize(event):
    print("widget", event.widget)
    print("height", event.height, "width", event.width)
    
def load_graph():
    global graph
    global order
    global order_reverse
    graph = read_graph(sys.argv[1])
    # draw_graph(graph)

def _create_circle(self, x, y, r, **kwargs):
    return self.create_oval(x-r, y-r, x+r, y+r, **kwargs)

Canvas.create_circle = _create_circle

def draw_graph_force():
    canvas.delete('all')
    hrs = RECT_SIZE/2.0
    pos = nx.spring_layout(graph.G)

    minx, miny, maxx, maxy = float('inf'), float('inf'), float('-inf'), float('-inf')
    for v in graph.G.nodes:
        if pos[v][0] > maxx:
            maxx = pos[v][0]
        if pos[v][0] < minx:
            minx = pos[v][0]
        if pos[v][1] > maxy:
            maxy = pos[v][1]
        if pos[v][1] < miny:
            miny = pos[v][1]

    newpos = {}

    for p in pos:
        newpos[p] = ((pos[p][0]-minx)*(300/(maxx-minx))+50, (pos[p][1]-miny)*(300/(maxy-miny))+50)

    for e in graph.G.edges:
        s = tuple(newpos[e[0]])
        t = tuple(newpos[e[1]])

        # print(e, s, t)

        canvas.create_line(*s, *t, fill="#AAA", width=w+0.5)
        canvas.create_line(*s, *t, width=w)

    for v in graph.G.nodes:
        if v < graph.a:
            color = "#9fc0de"
        else:
            color = "#f2c894"
        canvas.create_circle(*newpos[v], hrs, width=w+0.5, outline="#AAA")
        canvas.create_circle(*newpos[v], hrs, width=w, fill=color)
        canvas.create_text(*newpos[v], text=f"{v}", fill="black")

def reorder():
    global order
    global order_reverse
    order = {}
    with open(sys.argv[2]) as file:
        lines = file.readlines()
        counter = 0
        for line in lines:
            order[int(line)] = counter
            order_reverse[counter] = int(line)
            counter += 1
    print(order)
    print(order_reverse)


def draw_graph_with_order():
    canvas.delete('all')
    hrs = RECT_SIZE/2.0
    lcenters = [(CPAD + hrs + (RPAD + RECT_SIZE + HDIST) * i, CPAD + hrs) for i in range(0, graph.a)]
    rcenters = [(CPAD + hrs + (RPAD + RECT_SIZE + HDIST) * i, CPAD + hrs + LDIST) for i in range(0, graph.b)]
        
    for e in graph.G.edges:
        s = lcenters[e[0]]
        t = rcenters[order[e[1]]]
        
        canvas.create_line(*s, *t, fill="#AAA", width=w+0.5)
        canvas.create_line(*s, *t, width=w)
        
    for i, c in enumerate(lcenters + rcenters):
        # canvas.create_rectangle(c[0] - hrs, c[1] - hrs, c[0] + hrs, c[1] + hrs, fill="black")
        if i < graph.a:
            color = "#9fc0de"
            canvas.create_circle(c[0], c[1], hrs, width=w+0.5, outline="#AAA")
            canvas.create_circle(c[0], c[1], hrs, width=w, fill=color)
            canvas.create_text(c, text=f"{i}", fill="black")
        else:
            color = "#f2c894"
            canvas.create_circle(c[0], c[1], hrs, width=w+0.5, outline="#AAA")
            canvas.create_circle(c[0], c[1], hrs, width=w, fill=color)
            canvas.create_text(c, text=f"{order_reverse[i-graph.a]}", fill="black")

def draw_graph():
    canvas.delete('all')
    hrs = RECT_SIZE/2.0
    lcenters = [(CPAD + hrs + (RPAD + RECT_SIZE + HDIST) * i, CPAD + hrs) for i in range(0, graph.a)]
    rcenters = [(CPAD + hrs + (RPAD + RECT_SIZE + HDIST) * i, CPAD + hrs + LDIST) for i in range(0, graph.b)]
        
    for e in graph.G.edges:
        s = lcenters[e[0]]
        t = rcenters[e[1] - graph.a]
        
        canvas.create_line(*s, *t, fill="#AAA", width=w+0.5)
        canvas.create_line(*s, *t, width=w)
        
    for i, c in enumerate(lcenters + rcenters):
        # canvas.create_rectangle(c[0] - hrs, c[1] - hrs, c[0] + hrs, c[1] + hrs, fill="black")
        if i < graph.a:
            color = "#9fc0de"
        else:
            color = "#f2c894"
        canvas.create_circle(c[0], c[1], hrs, width=w+0.5, outline="#AAA")
        canvas.create_circle(c[0], c[1], hrs, width=w, fill=color)
        canvas.create_text(c, text=f"{i}", fill="black")

# def select_file():
#     filetypes = (('instance files', '*.gr'))
#     filename = fd.askopenfilename(title='Open a file', initialdir='/', filetypes=filetypes)
#     showinfo(title='Selected File', message=filename)

def visualize():
    root = Tk()
    root.title("Pace 2024")

    mainframe = ttk.Frame(root, width=600, height=400, padding="3 3 12 12")
    mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    mainframe.rowconfigure(0, weight=1)
    mainframe.columnconfigure(0, weight=1)

    global canvas 
    canvas = Canvas(mainframe, width=600, height=400, bg="white")
    canvas.grid(column=0, row=0, sticky=W, columnspan=5)
    canvas.rowconfigure(0, weight=1)
    canvas.columnconfigure(0, weight=1)

    loadbutton = ttk.Button(mainframe, text="Load graph", command=load_graph)
    loadbutton.grid(column=0, row=1, sticky=W)
    loadbutton.rowconfigure(1, weight=1)
    loadbutton.columnconfigure(0, weight=1)

    orderbutton = ttk.Button(mainframe, text="Load solution", command=reorder)
    orderbutton.grid(column=1, row=1, sticky=W)
    orderbutton.rowconfigure(1, weight=1)
    orderbutton.columnconfigure(1, weight=1)

    drawbutton = ttk.Button(mainframe, text="Draw 2-sided", command=draw_graph)
    drawbutton.grid(column=2, row=1, sticky=W)
    drawbutton.rowconfigure(1, weight=1)
    drawbutton.columnconfigure(2, weight=1)

    drawsolbutton = ttk.Button(mainframe, text="Draw 2-sided solved", command=draw_graph_with_order)
    drawsolbutton.grid(column=3, row=1, sticky=W)
    drawsolbutton.rowconfigure(1, weight=1)
    drawsolbutton.columnconfigure(3, weight=1)

    drawforcebutton = ttk.Button(mainframe, text="Draw force", command=draw_graph_force)
    drawforcebutton.grid(column=4, row=1, sticky=W)
    drawforcebutton.rowconfigure(1, weight=1)
    drawforcebutton.columnconfigure(4, weight=1)

    # canvas.bind("<Configure>", resize)
    load_graph()
    reorder()

    root.mainloop()


from typing import TypeVar, TextIO, Tuple, List, Dict
import pathlib
import segment_tree as sg

PG = TypeVar('PG', bound='PaceGraph')

class PaceGraph:
    # We assume only the right side is flexible
    right: list
    right_order: Dict[int, int]
    left: int

    edgeset: List[Tuple[int, int]]
    edges: Dict[int, List[int]]

    # Build a graph with sides A and B of size a and b. The side B is assumed to be flexible and A to be fixed.
    # If order is given we set the right side to be ordered that way. 
    def __init__(self, a: int, b: int, edges: list = None, order: list = None):
        if order:
            oelements = set(order)
            relements = set(range(a + 1, a + b + 1))
            diff = oelements - relements
            if len(diff) != 0 :
                raise ValueError(f"Error! 'order' contains elements not present in the graph {diff}.")
            else:
                self.right = order
        else:
            self.right = list(range(a + 1, a + b + 1))
        
        self.right_order = {u: p for p, u in enumerate(self.right)}
        self.left = a

        self.edges = {}
        self.edgeset = []
        if edges:
            edges = sorted(edges) # sort all edges such that we get sorted adjacency lists
            for e in edges:
                u = e[0] if e[0] < e[1] else e[1]
                v = e[0] if e[1] < e[0] else e[1]

                if not (u in self.edges.keys()):
                    self.edges[u] = []

                if not (v in self.edges.keys()):
                    self.edges[v] = []

                self.edges[u].append(v)
                self.edges[v].append(u)
                self.edgeset.append((u,v))

    # Return the neighbors of the given vertex
    def neighbors(self, u: int) -> list[int]:
        return self.edges[u] if u in self.edges.keys() else []
    
    # Set the order to the given list
    def set_order(self, order: List[int], check: bool = False):
        if check:
            oelements = set(order)
            relements = set(range(len(self.right) + 1, len(self.right) + self.left + 1))
            diff = oelements - relements
            if len(diff) != 0 :
                raise ValueError(f"Error! 'order' contains elements not present in the graph {diff}.")
            else:
                self.right = order

        self.right = order
        self.right_order = {u: p for p, u in enumerate(self.right)}

    # Check if edges (a,b) and (c,d) cross
    def cross(self, a: int, b: int, c: int, d: int) -> bool:
        if a > b:
            a, b = b, a
            
        if c > d:
            c, d = d, c

        b = self.right_order[b]
        d = self.right_order[d]
        
        return (a < c and b > d) or (c < a and d > b)
    
    # Swap the position of two vertices in the order
    def swap(self, a: int, b: int):
        apos = self.right_order[a]
        bpos = self.right_order[b]

        self.right[apos], self.right[bpos] = self.right[bpos], self.right[apos]
        self.right_order[a] = bpos
        self.right_order[b] = apos

    # Count the crossings in the trivial way by iterating all edges and checking if they cross.
    def countcrossings_trivial(self) -> int:
        crossings = 0
        for a in range(1, self.left+1):
            for c in range(a + 1, self.left+1):
                for b in self.neighbors(a):
                    for d in self.neighbors(c):
                        if self.cross(a, b, c, d):
                            crossings += 1

        return int(crossings)
    
    # Count the crossings using a stack-like implementation.
    def countcrossings_stacklike(self) -> int:
        leftside = sorted(self.edgeset, key=lambda e: (e[0], self.right_order[e[1]]), reverse=True)
        rightside = sorted(self.edgeset, key=lambda e: (self.right_order[e[1]], e[0]), reverse=True)
        leftpos = {e: p for p, e in enumerate(leftside)}
        
        openedges = set()
        doneedges = set()
        crossings = 0
        for i in range(0, len(rightside)):
            e = rightside[i]
            
            if e in openedges:
                openedges.remove(e)
                
            for f in leftside[i:leftpos[e]]:
                if f not in doneedges:
                    openedges.add(f)
            
            ecr = 0
            for f in openedges:
                if self.cross(*e, *f):
                    ecr += 1
                    
            crossings += ecr
            doneedges.add(e)
        
        return crossings
    
    def countcrossings_segtree(self) -> int:
        size = len(self.right)
        edges = sorted(self.edgeset, key=lambda e: (e[0], self.right_order[e[1]]))

        arr = [0] * (size + 1)
        t = sg.SegmentTree(arr)

        crossings = 0
        for edge in edges:
            old = t.query(self.right_order[edge[1]], self.right_order[edge[1]], "sum")
            t.update(self.right_order[edge[1]], old+1)

            if size > self.right_order[edge[1]]+1:
                crossings_found = t.query(self.right_order[edge[1]]+1, size, "sum")
                crossings += crossings_found

        return crossings
            

    # Create a graph from a io object containing gr-formated graph
    @classmethod
    def from_gr(cls, gr: TextIO, order: list = None) -> PG:
        a = 0
        b = 0
        pfound = False
        edges = []
        for line in gr:
            if line[0] == "p":
                a, b = list(map(int, line.split(" ")[2:4]))
                pfound = True
            elif line[0] == "c":
                pass
            elif pfound:
                edges.append(tuple(map(int, line.split(" "))))
            else:
                raise ValueError("ERROR: Encountered edge before p-line.")
                
        return PaceGraph(a, b, edges, order)

    # Return the graph in gr format
    def to_gr(self) -> str:
        newline = "\n"
        thestring = f"p ocr {self.left} {len(self.right)} {len(self.edgeset)}\n"
        thestring = thestring + newline.join([f"{e[0]} {e[1]}" for e in self.edgeset])

        return thestring

    # Return the graph with cutwidth in gr format
    def to_gr_cutwidth(self, permutation, cutwidth) -> str:
        newline = "\n"
        thestring = f"p ocr {self.left} {len(self.right)} {len(self.edgeset)} {cutwidth}\n"
        thestring = thestring + newline.join([f"{i}" for i in permutation])
        thestring = thestring + newline.join([f"{e[0]} {e[1]}" for e in self.edgeset])

        return thestring
            
def write_graph(graph: PaceGraph, filename: str):
    with open(filename, "w", encoding='utf-8') as f:
        f.write(graph.to_gr())

def write_graph_cutwidth(graph: PaceGraph, permutation: list, cutwidth: int, filename: str):
    with open(filename, "w") as f:
        f.write(graph.to_gr_cutwidth(permutation, cutwidth))
        
def write_solution(order: list, filename: str):
    with open(filename, "w", encoding='utf-8') as f:
        for n in order:
            f.write(f"{n}\n")

def read_graph(filename: pathlib.Path, order: list = None) -> PaceGraph:
    with open(filename, "r", encoding='utf-8') as f:
        return PaceGraph.from_gr(f, order)

def read_solution(filename: str) -> list:
    order = []
    with open(filename, "r", encoding='utf-8') as f:
        for line in f.readlines():
            order.append(int(line.strip()))
            
    return order


import click
import pathlib
import subprocess
import tempfile
import importlib.resources as irs
from .pace import read_graph, read_solution

@click.command()
@click.option('--interleave', 'method', flag_value='interleave', help='Count crossings checking pairs of edges one by one.')
@click.option('--stacklike', 'method', flag_value='stacklike', help='Count crossings using a stack-like method, similar to counting crossings in book drawings.')
@click.option('--segtree', 'method', flag_value='segtree', default=True, help='Count crossings using a segment tree. [default]')
@click.option('-c', '--only-crossings', is_flag=True, default=False, help='Print only the found number of crossings.')
@click.argument('graph', type=click.Path(exists=True, path_type=pathlib.Path))
@click.argument('solution', type=click.Path(exists=True, path_type=pathlib.Path))
def verify(method, only_crossings, graph, solution):
    '''Print the number of crossings the given solution has.'''
    solution = read_solution(solution)
    graph = read_graph(graph, solution)

    if method == "interleave":
        crossings = graph.countcrossings_trivial()
    elif method == "stacklike":
        crossings = graph.countcrossings_stacklike()
    else:
        crossings = graph.countcrossings_segtree()

    if only_crossings:
        click.echo(crossings)
    else:
        click.echo(f"Using {solution} as ordering of the vertices we found {crossings} crossings using the {method} method.")

def solutionfromstdout(stdout):
    return list(map(int, stdout.strip().split("\n")))

def print_result(control: int, test: int):
    if control == test:
        click.echo(f"The solver and the solution are the same!")
    else:
        click.echo(f"The solver and the solution are NOT the same!")

    click.echo(f"Solver: {test} Solution: {control}")
    click.echo()

def print_test(instance: pathlib.Path):
    click.echo(f"Testing now {instance[0].name}...")

@click.command()
@click.option('--tiny/--no-tiny', default=True, help="If set run the tests on the tiny test set.")
@click.option('--test', type=click.Path(exists=True, path_type=pathlib.Path, file_okay=False), multiple=True, default=None, help="Add additional test sets. Can be supplied multiple times of directories, each of them should contain a folder instances with .gr files and a folder solutions containing .sol files with the same file names else.")
@click.option('--instanceas', type=click.Choice(['file', 'stdin']), default='file', help="How the instance is passed to the solver, either as a path to a file or on stdin.")
@click.option('--solutionas', type=click.Choice(['file', 'stdout']), default='file', help="How the soltution is returned by the solver. Either written to a file, we provide a temporary file path, or to stdout.")
@click.option('-c', '--only-compare', is_flag=True, default=False, help='Print only the number of crossings by the solver and in the solution.')
@click.argument('solver', type=str)
def test(tiny, test, instanceas, solutionas, only_compare, solver):
    '''Test the given solver on the provided test instances.'''

    instances = [] # Collect the set of instances as a list of tuples

    def getsortedfiles(directory):
        return zip(sorted(directory.joinpath("instances").iterdir()), sorted(directory.joinpath("solutions").iterdir()))

    if tiny:
        tinytests = irs.files("pace2024_verifier.tiny_test_set")
        instances += list(getsortedfiles(tinytests))

    if test:
        for testdir in test:
            instances += list(getsortedfiles(testdir))

    if solutionas == 'file':
        with tempfile.TemporaryDirectory() as tmp:        
            for instance in instances:
                print_test(instance)

                tmpsolution = pathlib.Path(tmp).joinpath(f"{instance[1].stem}.sol")
                testsolution = None

                if instanceas == 'file':
                    ret = subprocess.run([solver, instance[0], tmpsolution], text=True, capture_output=True)
                    testsolution = read_solution(tmpsolution)
                else:
                    graph = read_graph(instance[0])
                    ret = subprocess.run([solver, tmpsolution], input=graph.to_gr(), text=True, capture_output=True)
                    testsolution = read_solution(tmpsolution)

                controlsolution = read_solution(instance[1])

                graph = read_graph(instance[0], controlsolution)
                controlcrossings = graph.countcrossings_segtree()
                graph.set_order(testsolution)
                testcrossings = graph.countcrossings_segtree()

                if not only_compare:
                    print_result(controlcrossings, testcrossings)
                else:
                    click.echo(f"{testcrossings},{controlcrossings}")
    else:
        for instance in instances:
            print_test(instance)

            testsolution = None

            if instanceas == 'file':
                ret = subprocess.run([solver, instance[0]], text=True, capture_output=True)
                testsolution = solutionfromstdout(ret.stdout)
            else:
                graph = read_graph(instance[0])
                ret = subprocess.run([solver], input=graph.to_gr(), text=True, capture_output=True)
                testsolution = solutionfromstdout(ret.stdout)

            controlsolution = read_solution(instance[1])

            graph = read_graph(instance[0], controlsolution)
            controlcrossings = graph.countcrossings_segtree()
            graph.set_order(testsolution)
            testcrossings = graph.countcrossings_segtree()

            if not only_compare:
                print_result(controlcrossings, testcrossings)
            else:
                click.echo(f"{testcrossings},{controlcrossings}")
            
